#!/usr/bin/env python3

################################################
#
#
#
################################################

################################################
#   Libraries
################################################
import json
from pathlib import Path
from typing import Any, Dict, Sequence
from packaging import version

from dcicutils import ff_utils


JsonObject = Dict[str, Any]

SMAHT_KEYS_FILE = Path.expanduser(Path("~/.smaht-keys.json")).absolute()


################################################
#   Functions
################################################
def make_embed_request(ids, fields, auth_key, single_item=False):
    """POST to embed API for retrieval of specified fields for given
    identifiers (from Postgres, not ES).

    :param ids: Item identifier(s)
    :type ids: str or list(str)
    :param fields: Fields to retrieve for identifiers
    :type fields: str or list(str)
    :param auth_key: Portal authorization key
    :type auth_key: dict
    :param single_item: Whether to return non-list result because only
         maximum one response is expected
    :type single_item: bool
    :return: Embed API response
    :rtype: list or dict or None
    """
    result = []
    if isinstance(ids, str):
        ids = [ids]
    if isinstance(fields, str):
        fields = [fields]
    id_chunks = chunk_ids(ids)
    server = auth_key.get("server")
    for id_chunk in id_chunks:
        post_body = {"ids": id_chunk, "fields": fields}
        embed_request = ff_utils.authorized_request(
            server + "/embed", verb="POST", auth=auth_key, data=json.dumps(post_body)
        ).json()
        result += embed_request
    if single_item:
        if not result:
            result = None
        elif len(result) == 1:
            result = result[0]
        else:
            raise ValueError(
                "Expected at most a single response but received multiple: %s" % result
            )
    return result


def chunk_ids(ids):
    """Split list into list of lists of maximum chunk size length.

    Embed API currently accepts max 5 identifiers, so chunk size is 5.

    :param ids: Identifiers to chunk
    :type ids: list
    :return: Chunked identifiers
    :rtype: list
    """
    result = []
    chunk_size = 5
    for idx in range(0, len(ids), chunk_size):
        result.append(ids[idx: idx + chunk_size])
    return result


def check_status(meta_workflow_run, valid_final_status=None):
    """Check if MetaWorkflowRun status is valid.

    If given valid final status, check MetaWorkflowRun.final_status
    as well.

    :param meta_workflow_run: MetaWorkflowRun[json]
    :type meta_workflow_run: dict
    :param valid_status: Final status considered valid
    :type valid_status: list
    :return: Whether MetaWorkflowRun final_status is valid
    :rtype: bool
    """
    item_status = meta_workflow_run.get("status", "deleted")
    if item_status not in ["obsolete", "deleted"]:
        result = True
        if valid_final_status:
            final_status = meta_workflow_run.get("final_status")
            if final_status not in valid_final_status:
                result = False
    else:
        result = False
    return result


class AuthorizationError(Exception):
    pass


def get_cgap_keys_path() -> Path:
    return SMAHT_KEYS_FILE


# TODO: dcicutils.creds_utils handles all of this
def get_auth_key(env_key: str) -> JsonObject:
    keys_path = get_cgap_keys_path()
    with keys_path.open() as file_handle:
        keys = json.load(file_handle)
    key = keys.get(env_key)
    if key is None:
        raise AuthorizationError(
            f"No key in {str(SMAHT_KEYS_FILE.absolute())} matches '{env_key}'"
        )
    return key


def keep_last_item(items: Sequence) -> Sequence:
    if len(items) <= 1:
        result = items
    elif len(items) > 1:
        result = items[-1:]
    return result


def get_file_set(fileset_accession, smaht_key):
    """Get the fileset from its accession

    Args:
        fileset_accession (str): fileset accession
        smaht_key (dict): SMaHT key

    Returns:
        dict: Fileset item from portal
    """
    return ff_utils.get_metadata(
        fileset_accession, add_on="frame=raw&datastore=database", key=smaht_key
    )


def get_library_from_file_set(file_set, smaht_key):
    """Get the library that is associated with a fileset

    Args:
        file_set(dicr): fileset from portal
        smaht_key (dict): SMaHT key

    Raises:
        Exception: Raises an exception when there are multiple libraries associated

    Returns:
        dict: Library item from portal
    """

    if len(file_set["libraries"]) > 1:
        raise Exception(f"Multiple libraries found for fileset {file_set['accession']}")
    library = ff_utils.get_metadata(
        file_set["libraries"][0], add_on="frame=raw&datastore=database", key=smaht_key
    )
    return library

def get_sample_from_library(library, smaht_key):
    """Get the sample that is associated with a library

    Args:
        library (dict): library item from portal
        smaht_key (dict): SMaHT key

    Raises:
        Exception: Raises an exception when there are multiple samples associated

    Returns:
        dict: Sample item from portal
    """
    analyte = ff_utils.get_metadata(
        library["analyte"], add_on="frame=raw&datastore=database", key=smaht_key
    )
    if len(analyte["samples"]) > 1:
        raise Exception(f"Multiple samples found for library {library['accession']}")
    sample = ff_utils.get_metadata(
        analyte["samples"][0], add_on="frame=raw&datastore=database", key=smaht_key
    )
    return sample


def get_latest_mwf(mwf_name, smaht_key):
    """Get the latest version of the MWF with name `mwf_name`

    Args:
        mwf_name (string): Name of the MWF
        smaht_key (dcit): SMaHT key

    Returns:
        dict: MWF item from portal
    """
    query = f"/search/?type=MetaWorkflow&name={mwf_name}"
    search_results = ff_utils.search_metadata(query, key=smaht_key)
    
    if len(search_results) == 0:
        return None
    
    latest_result = search_results[0]
    if len(search_results) == 1:
        return latest_result
    
    # There are multiple MWFs. Get the latest version
    for search_result in search_results:
        if version.parse(latest_result["version"]) < version.parse(search_result["version"]):
            latest_result = search_result
    return latest_result


def get_mwfr_file_input_arg(argument_name, files):
    return {"argument_name": argument_name, "argument_type": "file", "files": files}


def get_mwfr_parameter_input_arg(argument_name, value):
    return {
        "argument_name": argument_name,
        "argument_type": "parameter",
        "value": value,
    }



#!/usr/bin/env python3

################################################
#
#
#
################################################

################################################
#   Libraries
################################################
import sys, os

# dcicutils
from dcicutils import ff_utils
from dcicutils.s3_utils import s3Utils
from tibanna.job import Job


################################################
#   FFWfrUtils
################################################
class FFWfrUtils(object):
    def __init__(self, env):
        """
        :param env: e.g. 'smaht-wolf', 'smaht-wolf'
        :type env: str
        """
        self.env = env

        # Cache for metadata
        self._metadata = dict()
        # Cache for access key
        self._ff_key = None

    def wfr_run_uuid(self, job_id):
        """This is the function to be used by Magma.
        """
        wfr_meta = self.wfr_metadata(job_id)
        if not wfr_meta:
            return None
        return wfr_meta['uuid']

    def wfr_run_status(self, job_id):
        """This is the function to be used by Magma.
        Return the status of the run associated with specified job_id.
        If run associated with job_id is not found, we consider it failed.
        """
        wfr_meta = self.wfr_metadata(job_id)
        if not wfr_meta:
            return 'error'
        else:
            return wfr_meta['run_status']

    def get_minimal_processed_output(self, job_id):
        """This is the function to be used by Magma.
        Return a list of {'argument_name': <arg_name>, 'file': <uuid>}
        for all processed file output. If no output, return None.
        """
        wfr_output = self.wfr_output(job_id)
        return self.filter_wfr_output_minimal_processed(wfr_output)

    def wfr_output(self, job_id):
        """Return the raw output from the run metadata.
        Return None if run metadata associated with the job is not found or
        does not have associated output files.
        """
        if self.wfr_metadata(job_id):
            return self.wfr_metadata(job_id).get('output_files', None)
        else:
            return None

    def wfr_metadata(self, job_id):
        """Get portal run metadata from job_id.
        Return None if a run associated with job id cannot be found.
        """
        # Use cache
        if job_id in self._metadata:
            return self._metadata[job_id]
        # Search by job id
        query='/search/?type=WorkflowRun&job_id=%s' % job_id
        try:
            search_res = ff_utils.search_metadata(query, key=self.ff_key)
        except Exception as e:
            raise FdnConnectionException(e)
        if search_res:
            self._metadata[job_id] = search_res[0]
            return self._metadata[job_id]
        else:
            # find it from dynamoDB
            job_info = Job.info(job_id)
            if not job_info:
                return None
            wfr_uuid = job_info.get('WorkflowRun uuid', '')
            if not wfr_uuid:
                return None
            self._metadata[job_id] = ff_utils.get_metadata(wfr_uuid, add_on="frame=raw&datastore=database", key=self.ff_key)
            return self._metadata[job_id]
        
    @property
    def ff_key(self):
        """Get access key for the portal.
        """
        # Use cache
        if not self._ff_key:
            # Use tibanna key for now
            self._ff_key = s3Utils(env=self.env).get_access_keys('access_key_tibanna')
        return self._ff_key

    @staticmethod
    def filter_wfr_output_minimal_processed(wfr_output):
        """Return a list of {'argument_name': <arg_name>, 'file': <uuid>}
        for all processed file output.
        """
        if wfr_output:
            return [{'argument_name': opf['workflow_argument_name'],
                     'file': opf['value']['uuid']} \
                        for opf in wfr_output]
        else:
            return None

#end class


class FdnConnectionException(Exception):
    pass

#end class



################################################
#
#   Function to run MetaWorkflowRun[portal]
#       with tibanna and patch metadata
#
################################################
from dcicutils import ff_utils
from tibanna_ffcommon.core import API

from magma_ff import inputgenerator as ingen
from magma_ff.metawfl import MetaWorkflow
from magma_ff.metawflrun import MetaWorkflowRun
from magma_ff.utils import check_status, make_embed_request


################################################
#   Functions
################################################
def run_metawfr(
    metawfr_uuid,
    ff_key,
    verbose=False,
    sfn="tibanna_zebra",
    env="fourfront-cgap",
    maxcount=None,
    valid_status=None,
):
    """Launch pending runs on MetaWorkflowRun[portal] via tibanna.
    PATCH MetaWorkflowRun[portal] with updates.

    Can double-check MetaWorkflowRun.final_status is valid since
    grabbing item from Postgres here.

    :param metawfr_uuid: MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param sfn: Step function name
    :type sfn: str
    :param env: Environment name
    :type env: str
    :param maxcount: Maximum number of WorkflowRuns to create for the
        MetaWorkflowRun
    :type maxcount: int
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param valid_status: Status considered valid for MetaWorkflowRun
        final_status property
    :type valid_status: list(str) or None
    """
    perform_action = True
    embed_fields = ["*", "meta_workflow.*"]
    meta_workflow_run = make_embed_request(
        metawfr_uuid, embed_fields, ff_key, single_item=True
    )
    meta_workflow = meta_workflow_run.get("meta_workflow")
    perform_action = check_status(meta_workflow_run, valid_status)
    if perform_action:
        run_obj = MetaWorkflowRun(meta_workflow_run)
        wfl_obj = MetaWorkflow(meta_workflow)
        ingen_obj = ingen.InputGenerator(wfl_obj, run_obj)
        in_gen = ingen_obj.input_generator(env)
        count = 0
        for input_json, patch_dict in in_gen:
            API().run_workflow(input_json=input_json, sfn=sfn)  # Start tibanna run
            res_post = ff_utils.patch_metadata(patch_dict, metawfr_uuid, key=ff_key)
            if verbose:
                print(res_post)
            count += 1
            if maxcount and count >= maxcount:
                break
